/*
 * Copyright 2008 Evan Cowden
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
package org.nakedosgi.processor.scr;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.annotation.processing.Processor;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.TypeMirror;
import javax.xml.bind.JAXBException;
import org.nakedosgi.annotations.NullableBoolean;
import org.nakedosgi.annotations.scr.SCRComponent;
import org.nakedosgi.annotations.scr.SCRReference;
import org.nakedosgi.processor.OSGiProcessor;
import org.nakedosgi.processor.ParsingUtil;
import org.nakedosgi.processor.scr.model.Component;
import org.nakedosgi.processor.scr.model.Implementation;
import org.nakedosgi.processor.scr.model.Provide;
import org.nakedosgi.processor.scr.model.Reference;
import org.nakedosgi.processor.scr.model.Service;
import org.nakedosgi.spi.ServiceProvider;
import static org.nakedosgi.Constants.*;

/**
 * Processor for the {@link SCRComponent} annotation
 * @author Evan Cowden
 */
@ServiceProvider(Processor.class)
@SupportedSourceVersion(SourceVersion.RELEASE_6)
@SupportedAnnotationTypes("org.nakedosgi.annotations.scr.SCRComponent")
public class SCRComponentProcessor extends OSGiProcessor {

    //Map between component classes and components
    private Map<String, Component> components = SCRProcessingUtil.getComponents();

    @Override
    public boolean processBundle(Set<? extends TypeElement> annotations, RoundEnvironment roundEnv) {
        Set<? extends Element> elements = roundEnv.getElementsAnnotatedWith(SCRComponent.class);

        for (Element element : elements) {
            Component result = new Component();

            //Get info
            TypeElement type = (TypeElement) element; //Safe since SCRComponent has a @Target restraint
            SCRComponent annotation = type.getAnnotation(SCRComponent.class);

            //Parse top-level component info
            //...Implementation class
            Implementation implementation = new Implementation();
            String qualifiedName = ParsingUtil.getQualifiedName(type.asType());
            implementation.setClazz(qualifiedName);
            result.setImplementation(implementation);
            
            //...enabled
            result.setEnabled(annotation.enabled());

            //...name
            if (annotation.name().equals(NO_VALUE)) {
                result.setName(type.getSimpleName().toString());
            } else {
                result.setName(annotation.name());
            }

            //...factory
            if (!annotation.factory().equals(NO_VALUE)) {
                result.setFactory(annotation.factory());
            }

            //...immediate
            if (!annotation.immediate().equals(NullableBoolean.UNSPECIFIED)) {
                result.setImmediate(annotation.immediate().value());
            }

            //...serviceFactory
            if (annotation.serviceFactory() == true) {
                //...create new Service if one does not exist
                Service service = result.getService();
                if (service == null) {
                    service = new Service();
                    result.setService(service);
                }

                service.setServicefactory(annotation.serviceFactory());
            }

            //...interfaces
            List<String> interfaces = new ArrayList();  //Find all interfaces, put them here, then add to component model later
            //......Find all interfaces
            if (annotation.interfaces().length == 0) {//TODO might actually want none...
                //...find declared interfaces
                List<? extends TypeMirror> declaredInterfaces = type.getInterfaces();
                for (TypeMirror typeMirror : declaredInterfaces) {
                    interfaces.add(typeMirror.toString());
                }
            } else {
                //...find interfaces in annotation
                for (Class iface : annotation.interfaces()) {
                    //TODO can do more type checking here to confirm it is an interface, etc.
                    interfaces.add(iface.getName());
                }
            }
            //......then add to the component model
            if (interfaces.size() > 0) {
                Service service = result.getService();
                if (service == null) {
                    service = new Service();
                    result.setService(service);
                }
                for (String iface : interfaces) {
                    Provide provide = new Provide();
                    provide.setInterface(iface);
                    service.getProvide().add(provide);
                }
            }

            //...references
            for (SCRReference scrReference : annotation.references()) {
                Reference reference = new Reference(scrReference);
                result.getReference().add(reference);
            }

            String className = type.getQualifiedName().toString();
            components.put(className, result);
        }

        return true;
    }

    @Override
    public void processingDone() {
        //Write configuration files
        try {
            SCRProcessingUtil.writeComponentDescriptors(components.values(), processingEnv);
        } catch (JAXBException ex) {
            //TODO
            System.out.println("An error occured while writing configuration");
            ex.printStackTrace();
        } catch (IOException ex) {
            //TODO
            System.out.println("An error occured while writing configuration");
            ex.printStackTrace();
        }
        
        //Add component definitions to the manifest
        SCRProcessingUtil.addSCRComponents(manifest, components.values());
    }

    /**
     * Get the specified component out of the components collection.
     * If it does not already exist, create it and add it to the collection.
     * @param componentClass
     * @param components
     * @return
     */
    protected static Component getComponent(String componentClass, Map<String, Component> components) {
        Component result = components.get(componentClass);

        if (result == null) {
            result = new Component();
            components.put(componentClass, result);
        }

        return result;
    }
}
